// -*- C++ -*-

// CircularBuffer.h
//
// Circular buffer for aprl
// dpwe@media.mit.edu 1995feb02
//

#ifndef _CIRCULARBUFFER_H
#define _CIRCULARBUFFER_H

#include "Data.h"	// type definition of DATA (e.g. float)

class CircularBuffer {	// data stored in a circular buffer
private:
    DATA *buf;		// pointer to actual store
    int bufsize;	// total capacity of <buf>
    int head;		// offset of next location in buf to be written
    int tail;		// offset of next location in buf to be read
    int savedPos;	// for PushHead / PopTail
    int count;		// count of total elements put into buffer
    int checkingTail;	// boolean enables checking for overflow
    int isEmpty;	// if head==tail, are we empty or full?

    int _write_at(DATA *data, int n, int where)	// push in data, no checking
	{ int rem = bufsize - where;   
	  int part1 = min(rem,n);	int part2 = n-rem;
	  DATAcopy(data, buf+where, part1);
	  if(part2>=0)	{
	      DATAcopy(data+part1, buf, part2);
	      return part2;			 // the updated 'where'
	  } else {  return where+part1;	} }	 // no wrap occurred 
    int _read_at(DATA *data, int n, int where)   // pull out data, no checking
	{ int rem = bufsize - where;   
	  int part1 = min(rem,n);	int part2 = n-rem;
	  DATAcopy(buf+where, data, part1);  
	  if(part2>=0)	{
	      DATAcopy(buf, data+part1, part2);
	      return part2;			// the updated 'where'
	  } else {  return where+part1; } }	// no wrap occurred 

public:
    CircularBuffer(int size) : bufsize(size)  
        { if(bufsize)  { buf = new DATA[bufsize]; DATAclear(buf, bufsize); }
          savedPos = checkingTail = head = tail = count = 0; 
          isEmpty = 1; }
    ~CircularBuffer(void)
        { if(bufsize) delete [] buf; }

    void print(char *tag=0);
    int Resize(int size);	// Resize the buffer
    int Avai(void) const  { 
	                    int avai = (bufsize>0)?
				1+(bufsize-1+head-tail)%bufsize: 0;
			    if(head==tail && isEmpty) avai = 0;
			    return avai; }
	// if tail == head, then buffer is absolutely full.  Note that when
        // count==0, head==tail, but in fact we're empty.
    int Count(void) const
        { return count; }	// total number of data written
    int TailCount(void)	const	// absolute count corresponding to current tail
    	{ return count - Avai(); }
    int& CheckTail(void) { return checkingTail; }
    int Size(void) const  { return bufsize; }
    int GetPos(int base) const	// return an offset from buf of the data 
    			      	// whose absolute index is base - or -1 if 
    			     	// if it's no longer in the buffer
	{ // The buffer contains vals (max(0,count-bufsize)) to (count-1)
	  int earliest = max(0,count-bufsize);
	  // OK, it is in the buffer; must be (count-base) *before* head ptr
	  return (base < earliest || base >= count) ? -1:
	               ((bufsize+head - (count-base))%bufsize);  }
    void PushHead(void)	 { savedPos = count; }  // checkpoint for later restore
    void PushTail(void)	 { savedPos = TailCount(); }  // also for current tail
    int  PopTail(void)   { int newtail = GetPos(savedPos);
			   int retval = -1;
			   if(newtail>=0)  {
			       tail = newtail; retval=savedPos;
			       if(savedPos < count)	isEmpty = 0;
			   } return retval; }
    int UpdateHead(int n)    // after a pointer has been obtained with 
			     // GetHeadPtr, this function notifies the 
			     // buffer which new points have been written.
			     // The index of the next item (i.e. 'count') 
			     // is returned.
	{ head = (head+n)%bufsize;  if(n) isEmpty = 0;
	  count += n; return count; }
    int UpdateTail(int n)	// advance the tail pointer by n slots 
    				// after, presumably, reading them out of 
    				// a return from GetTailPtr
    				// Return absolute index of next tail point
	{ assert(n <= Avai());
	  tail = (tail+n)%bufsize;  if(n && tail==head) isEmpty = 1;
	  return TailCount(); }
    void Flush()		// bring the tail up to date
        { UpdateTail(Avai()); }
    int Write(DATA *data, int n)     // push data into buffer.  return n taken
        { if(checkingTail)  n = min(n, (tail-1-head+bufsize)%bufsize);
	  else		    n = min(n, bufsize);
	  _write_at(data, n, head); UpdateHead(n); return n; }
    int skip(int n)	   // drop samples, update tail
        { n = min(n, Avai());  UpdateTail(n); return n; }
    int Read(DATA *data, int n)  // get data from tail.  return actual n got
        { // I think you have to be checking tail if you're using get
	  n = min(n, Avai());
	  _read_at(data, n, tail); UpdateTail(n); return n; }
    int readWrite(DATA *in, DATA *out, int n)     // read and write the data
        { DATA *tmpout = out;
	  if(in==out)   tmpout = new DATA[n];
	  n = Read(tmpout, n);	Write(in, n);
	  if(in==out)	{ DATAcopy(tmpout, out, n); delete [] tmpout; }
	  return n; }
    int readFrom(DATA *data, int n, int from)  // access an absolute pos
        { int earliest = max(0, count-bufsize);
	  if (earliest > from || (from+n) > count)	
	      n = 0;	// just punt - a lose
	  else
	      _read_at(data, n, (bufsize+head-(count-from)) % bufsize);
	  return n; }
    // a rather different kind of operation returns pointers to areas
    int GetHeadPtr(DATA **pdata) // returns a head pointer and the number 
			         // of elements that can be written to it
        { int n = bufsize-head; 
          if(checkingTail && tail > head)  n = tail-head;
          *pdata = buf+head;  return n; }
    int GetTailPtr(DATA **pdata)  // returns a pointer to the next data to 
    				  // be read out, according to tail, and 
    				  // how many valid points there are there
	{ int n = bufsize-tail;
	  if(head > tail || (isEmpty && head==tail))	n = head-tail;
	  *pdata = buf+tail;	return n; }
    int GetTailPtrFrom(DATA **pdata, int base)  // returns a pointer to the 
					// requested history value and the 
					// number of valid values following
        { int basepos = GetPos(base);
	  int retval = 0;
          if(basepos >= 0)  {	// i.e. base is available
	      *pdata = buf + basepos;
	      int maxstraight = (basepos>=head)?bufsize-basepos:head-basepos;
	      retval = maxstraight;
	  }
	  return retval; }
    int SetTail(int base)		// Set the tail to abs. count base
    					// returns -1 if not avai.
	{ int newtail = GetPos(base);
	  if(newtail < 0)	base = -1;	// indicate fail by retval
	  else	{ 
	      tail = newtail;
	      if(base < count)	isEmpty = 0;
	  }
	  return base; }
    int PushZeros(int n) 		// act as if this many zeros
	                                // were just pushed
       { int todo = min(n,bufsize);  
	 int remain = todo;
	 DATA *d;
	 int m = min(remain, GetHeadPtr(&d));
	 DATAclear(d, m);	remain -= m;
	 UpdateHead(m);
	 if(n)  {  GetHeadPtr(&d); DATAclear(d, remain); UpdateHead(remain); }
	 return todo; }
    int PushVals(int n, DATA v) 	// act as if this many 'v's
	                                // were just pushed
       { int todo = min(n,bufsize);  
	 int remain = todo;
	 DATA *d;
	 int m = min(remain, GetHeadPtr(&d));
	 DATAset(d, v, m);	remain -= m;
	 UpdateHead(m);
	 if(n)  {  GetHeadPtr(&d); DATAset(d,v,remain); UpdateHead(remain); }
	 return todo; }

};

double CircBufInnerProd(CircularBuffer *cb1, int base1, 
		        CircularBuffer *cb2, int base2, int npoints);
    /* calculate the inner product (sum of products of elements)
       for the <npoints> from the two circular buffers, starting at 
       the specified indices. */

int CircBufInputOutputFrom(CircularBuffer *input, int cb1base, int cb1len, 
			   CircularBuffer *output, DATAprocess *proc);
    /* apply the DATAprocess to <cb1len> points of input taken from 
       circular buffer <input> starting at item <cb1base>.
       Return number of input samples used, may be less than requested. */

double CircBufInnerProd(CircularBuffer *cb1, 
		        CircularBuffer *cb2, int npoints);
    /* calculate the inner product (sum of products of elements)
       for the <npoints> from the two circular buffers, starting at 
       the actual tails.  Crash if <npoints> not avai on both. */

int CircBufInputOutput(CircularBuffer *input, CircularBuffer *output, 
		       DATAprocess *proc);
    /* apply the DATAprocess to all the Avai() points in input, write 
       to output.
       Return number of input samples used, may be less than requested. */

int CircBufSimpleIO(CircularBuffer *input, CircularBuffer *output, 
		    DATAsimpleFn *proc);
    /* Like InputOutput, but with fewer options */

int CircBufTransfer(CircularBuffer *input, CircularBuffer *output, int n);
    /* just transfer n points from input to output.  Return actual num */

#endif /* _CIRCULARBUFFER_H */
