// -*- C++ -*-

// EventPipe.h
//
// Classes for sources and sinks of data
// Pipe that does one-shotting on a thresholded signal
// *and* makes a callback on each leading edge of its output
//
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb02
//

#ifndef _EVENTPIPE_H
#define _EVENTPIPE_H

#include "DataPipe.h"

// function type for callback
typedef void cbFnType(double time, void *userdata);

class EventPipe : public DataPipe {
private:
    // config
    double holdoffTime;
    double pulseTime;
    // state
    double lastTrigTime;
    double lastVal;	// for getting leading edges across block bdries
    // callback
    cbFnType *cbFn;	// called with time and userdata
    void *cbFnData;

    int process(DATA *in, DATA *out, int n, double baseTime, double sr);
public:
    EventPipe(double offtime = 0, double ptime = 0, 
	      cbFnType *fn=0, void *fData=0, 
	      DataSource *in = 0)
	: holdoffTime(offtime), pulseTime(ptime), cbFn(fn), cbFnData(fData), 
          DataPipe(in)
	{ lastTrigTime = -999.9; lastVal = 0; };
    virtual ~EventPipe(void)
        { };
    void SetTimes(double off, double pulse)
        { holdoffTime = off; pulseTime = pulse; }
    void SetCBFn(cbFnType *fn, void *udata)
	{ cbFn = fn; cbFnData = udata; }
    int Update(double time);
};

#endif /* _EVENTPIPE_H */
