// -*- C++ -*-

// UserCmd.cc
//
// A class which passes data to a user-defined c function.
//

#include <stdio.h>
#include <malloc.h>
#include "assert.h"

#include "UserCmd.h"

UserCmd::UserCmd(char **pnames,DataSource **psrcs,int pninputs,
		 char *pcname,double srate, double buftime) 
    : names(pnames),ninputs(pninputs),cname(pcname),
      DataSource(srate,(int)(srate*buftime))
    {
      int i;
      double maxsr = 0;
      double maxbtime = 1e40;
      
      input = NULL;
      SetCfunc(pcname);
      SetInputs(psrcs,pninputs);
      
      if (srate == 0) { // fastest of the inputs
	for (i=0;i!=ninputs;i++)
	  if (input[i]->SampleRate() > maxsr)
	    maxsr = input[i]->SampleRate();
	SetSampleRate(maxsr);
	printf("userfunc '%s' sample rate defaulted to %.2lf\n",
	       cname,SampleRate());
      }
      if (buftime == 0) { // longest buffer
	for (i=0;i!=ninputs;i++)
	  if (input[i]->Size()/input[i]->SampleRate() < maxbtime)
	    maxbtime = input[i]->Size()/input[i]->SampleRate();
	Resize((int)(maxbtime * SampleRate()));
	printf("userfunc '%s' buffer size defaulted to %d (%.2lf sec)\n",
	       cname,Size(),Size()/SampleRate());
      }
    }

UserCmd::~UserCmd() {}

void UserCmd::SetInputs(DataSource **srcs,int n) {
  int i;
  if (input) { // first, release the old ones
    for (i=0;i!=ninputs;i++)
      input[i]->UnregisterSink((DataSink *)this);
    free(input);
  }
  input = (DataSource **)calloc(n+1,sizeof(DataSource *));
  for (i=0;i!=n;i++) {
    input[i]=srcs[i];
    double ipseq = srcs[i]->Sequence();
    if (Sequence() <= ipseq)
      SetSequence(1+ipseq);
    srcs[i]->RegisterSink((DataSink *)this);
  }
  input[i] = NULL;
}

void UserCmd::SetCfunc(char *pcname) {
  int i;
  extern int num_cfuncs;
  extern name_table cnames[];
  char *s;
   
  i=0;
  while(i<num_cfuncs && (s=&(cnames[i].name[0]),strcmp(pcname,s))) i++;
  if (i == num_cfuncs)
    fprintf(stderr,"Unknown c function name \"%s\".\n",pcname);
  else
    cfunc = cnames[i].func;
}

int UserCmd::Update(double /* time */) {
  // build the data into an array of UC_Parm structures
  // and pass it to the c function specified in cfunc.
  int i = 0,ct=0;
  double maxbtime = 0;
  
  UC_parm *parms,retval;
  while (input[i++])
    ct++;

  parms=(UC_parm *)calloc(ct,sizeof(UC_parm));

  for (i=0;i!=ct;i++) {
    int size;
    
    parms[i].name = (char *)malloc((strlen(names[i])+1)*sizeof(char));
    strcpy(parms[i].name,names[i]);
    input[i]->PopTail();
    size = input[i]->Avai();
    parms[i].length = size;
    parms[i].s = (DATA *)malloc(size * sizeof(DATA));
    input[i]->Read(parms[i].s,size);
    parms[i].samplerate = input[i]->SampleRate();
    if (parms[i].length / parms[i].samplerate > maxbtime )
      maxbtime = parms[i].length/parms[i].samplerate;
  }
  retval.samplerate = SampleRate();
  retval.length = (int)(maxbtime * retval.samplerate);
  (*cfunc)(parms,&retval,ct);

  free(parms);

  PushHead();
  Write(retval.s,retval.length);
  free(retval.s); // is this a bad thing?
  return retval.length;
}
  
