//
// tclAprlMisc.cc
//
// Miscillaneous pieces for the tcl interface
// for aprl - audio pattern recognition language
// dpwe@media.mit.edu 1995feb12
//

#include "tclAprlMisc.h"

// The global hash table for converting tcl names into DataSources
Tcl_HashTable datasrcTable;		// Convert sound names to ptrs
// (it's initialized in the Scheduler initialize code)

// Global that holds the last DataSrc created, for automatic chaining
DataSource *globalDataSource = 0;
// .. and its name
char *globalDSnameStr  = 0;

char *UniqueSPName(Tcl_Interp *interp)
{   /* return a unique name for a sched proc in a static array to be used */
    static char s[32];
    static int count = 0;
    int exists;
    Tcl_CmdInfo inf;

    do  {
	// create a new name of the form "proc17"
	sprintf(s, "%s%d", "proc", count++);
	// check to see if it exists in the command namespace
	exists = Tcl_GetCommandInfo(interp, s, &inf);
    } while(exists);
    return s;
}

int ConfigSpecified(char *val, int argc, char **argv)
{	/* for use in configure, quickly scan down a list of input 
	   flags to see if the specified arg is present.  
	   Only checks even-numbered fields.  You should 
	   include the leading "-" in the <val>.  Returns 
	   argv index holding *argument* (1..argc-1) if found, 
	   else 0 */
    int i,l;

    for(i=0; i<argc; i+=2)  {
	l = strlen(argv[i]);
	if(strncmp(val, argv[i], l)==0) 
	    return 1+i;
    }
    return 0;
}
	
int DSfromString(Tcl_Interp *interp, char **name, DataSource **rslt, 
		 int takeDefault)
{   /* Return the data source keyed by the name <name> in the lookup 
       table.  If no match and if <takeDefault>, return the 
       last created source from <globalDataSource> */
    
    DataSource *in = 0;
    if(*name && strcmp(*name, NONESTR)!=0)  {
	// lookup the input in the hash table
	Tcl_HashEntry *entryPtr;
	entryPtr = Tcl_FindHashEntry(&datasrcTable, *name);
	if(entryPtr == NULL)  {
	    Tcl_AppendResult(interp, "input \"", *name, 
			     "\" does not match any data sources.", NULL);
	    return TCL_ERROR;
	} 
	in = (DataSource *)Tcl_GetHashValue(entryPtr);
    }
    if(in == 0 && takeDefault)  {
	in = globalDataSource;	// use the default - the last
	if(*name)	ckfree(*name);
	*name = strdup(globalDSnameStr);
    }
    *rslt = in;
    return TCL_OK;
}
