// -------------------------------------------------------------------
// tkWaveSink.cc
//
// tk widget to display waveform streaming in from a DataPipe
// (based on tkWaveform)
// dpwe@mit.edu 1995feb08  part of APRL - audio pattern recog lang
//
#ifndef lint
static char rcsid[] = "$Header: /u/drspeech/src/tcltk/aprl/RCS/tkWaveSink.cc,v 1.1 1996/11/18 18:42:29 dpwe Exp dpwe $ @media.mit.edu Machine Listening Group";
#endif /* lint */

#include "tclAprlMisc.h"

// Widget data structure

class WaveSink;	// fwd declare

typedef struct {
    // standard fields
    Tk_Window	tkwin;
    Display	*display;
    Tcl_Interp	*interp;
    Tk_3DBorder bgBorder;	/* Used for drawing background. */
    XColor	*fgColor;
    XColor	*selColor;
    int 	borderWidth;
    int 	padX;
    int 	padY;
    int 	width, height;
    int 	relief;		// appearance of whole window
    // data for the drawing activities
    GC  	gc;
    int 	doubleBuffer;	// do we double buffer?
    int 	updatePending;
    int		lineWidth;	// pixel width of drawn lines
    // fields describing the content
    int		bufLen;		// how many points into the past do we hold?
    int		bufConts;	// how many valid points actually in there?
    float	*buf;		// buffer containing the data
    double	scale;		// scaling for data onto screen
    double	offset;		// add this to scaled value before display
    char	*input;		// Name of tclDataSource object
    WaveSink	*obj;		// the actual C++ object
} tkWaveSink;

// Define a sink to take the data

class WaveSink : public DataSink {
private:
    tkWaveSink *ws;
public:
    WaveSink(tkWaveSink *w, DataSource *in=0) : ws(w), DataSink(in)
        { };
    ~WaveSink(void)
        { };
    int Update(double time);
};

extern "C" {
extern char *strdup(const char *s1);	// difficult to get this declared

// static function prototypes

int WaveSink_Init(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
int wavesinkCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv);
static int  wavesinkConfigure(Tcl_Interp *interp, tkWaveSink *sw, 
			     int argc, char **argv, int flags);
static void wavesinkDestroy(char* clientData);
static void wavesinkDisplay(ClientData clientData);
static void wavesinkEventProc(ClientData clientData, XEvent *eventPtr);
static int  wavesinkWidgetCmd(ClientData clientData, Tcl_Interp * interp, 
			     int argc, char **argv);

}

// Argv parsing info

#define NONESTR "<none>"

static Tk_ConfigSpec wavesinkConfigSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background", 
     "#cdb79e", Tk_Offset(tkWaveSink, bgBorder), 0 }, 
    {TK_CONFIG_SYNONYM, "-bg", "background", NULL, NULL, 0, 0}, 
    {TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground", 
     "#701040", Tk_Offset(tkWaveSink, fgColor), 0 }, 
    {TK_CONFIG_COLOR, "-selColor", "selColor", "SelColor", 
     "#A0A0A0", Tk_Offset(tkWaveSink, selColor), 0 }, 
    {TK_CONFIG_SYNONYM, "-fg", "foreground", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"2", Tk_Offset(tkWaveSink, borderWidth), 0},
    {TK_CONFIG_SYNONYM, "-bw", "borderWidth", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-padX", "padX", "PadX",
	"2", Tk_Offset(tkWaveSink, padX), 0},
    {TK_CONFIG_PIXELS, "-padY", "padY", "PadY",
	"2", Tk_Offset(tkWaveSink, padY), 0},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"400", Tk_Offset(tkWaveSink, width), 0},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"150", Tk_Offset(tkWaveSink, height), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"raised", Tk_Offset(tkWaveSink, relief), 0},
    {TK_CONFIG_INT, "-dbl", "doubleBuffer", "DoubleBuffer",
	"1", Tk_Offset(tkWaveSink, doubleBuffer), 0},
    {TK_CONFIG_DOUBLE, "-offset", "offset", "Offset",
	"0.0", Tk_Offset(tkWaveSink, offset), 0},
    {TK_CONFIG_DOUBLE, "-scale", "scale", "Scale",
	"1.0", Tk_Offset(tkWaveSink, scale), 0},
    {TK_CONFIG_INT, "-buflen", "buflen", "Buflen",
	"200", Tk_Offset(tkWaveSink, bufLen), 0},
    {TK_CONFIG_INT, "-bufconts", "bufconts", "Bufconts",
	"0", Tk_Offset(tkWaveSink, bufConts), 0},
    {TK_CONFIG_STRING, "-input", "input", "Input",
	NONESTR, Tk_Offset(tkWaveSink, input), 0},
    {TK_CONFIG_INT, "-linewidth", "lineWidth", "LineWidth",
	"1", Tk_Offset(tkWaveSink, lineWidth), 0},
    {TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0}
};

int WaveSink_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tk_Window main = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "wavesink", wavesinkCmd, (ClientData) main,
	    (void (*)(ClientData)) NULL);
    return TCL_OK;
}

int wavesinkCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv)
{   /* Tcl command to instance a new wavesink widget */
    Tk_Window	main = (Tk_Window) clientData; 
    tkWaveSink 	*sw; 
    Tk_Window	tkwin;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be\"", 
			 argv[0], " pathName ?options?\"", NULL);
	return TCL_ERROR;
    }
    tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], NULL);
    if(tkwin == NULL)  {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "WaveSink");

    // Allocate and initialize the structure
    sw = (tkWaveSink *)ckalloc(sizeof(tkWaveSink));
    sw->tkwin = tkwin;
    sw->display = Tk_Display(tkwin);
    sw->interp  = interp;
    sw->bgBorder= NULL;
    sw->fgColor = NULL;
    sw->selColor = NULL;
    sw->gc = None;
    sw->doubleBuffer = 1;
    sw->updatePending = 0;

    sw->scale= 1.0;
    sw->bufLen = 0;
    sw->bufConts = 0;
    sw->buf = NULL;

    sw->obj   = 0;	// the object - none at present
    sw->input = 0;	// DataSource input string

    Tk_CreateEventHandler(sw->tkwin, ExposureMask|StructureNotifyMask,
			  wavesinkEventProc, (ClientData)sw);
    Tcl_CreateCommand(interp, Tk_PathName(sw->tkwin), wavesinkWidgetCmd, 
		      (ClientData)sw, (void (*)(ClientData))NULL);
    if(wavesinkConfigure(interp, sw, argc-2, argv+2, 0) != TCL_OK)  {
	Tk_DestroyWindow(sw->tkwin);
	return TCL_ERROR;
    }
    interp->result = Tk_PathName(sw->tkwin);
    return TCL_OK;
}

static void wavesinkGetWorkArea(tkWaveSink *sw, int *px, int *py, int *pw, int *ph)
{   /* return the area that WaveSink is actually drawn in */
    Tk_Window tkwin = sw->tkwin;

    *px = sw->padX;
    *pw = Tk_Width(tkwin) - 2*sw->padX;
    *py = sw->padY;
    *ph = Tk_Height(tkwin) - 2*sw->padY;
}

static int findOneNumber(char **p, double *ret)
{   /* passed a modifiable pointer to a string, skip any whitespace 
       and attempt to parse a float that comes next.  If success, 
       return 1 and write value into ret.  If hit end of string without 
       a number, return 0.  If string is not a number, return -1.
       <p> is modified ready for next call. */
    char   *r, *s=*p;
    double val;

    /* skip leading ws */
    s += strspn(s, " \n\r\t");
    if(*s=='\0')  {	/* hit end of string */
	return 0;
    }
    val = strtod(s, &r);
    if(r==s) {	/* no conversion was performed - bad string? */
	return -1;
    }
    *ret = val;
    *p = r;
    return 1;
}

static int wavesinkParseData(Tcl_Interp *interp, tkWaveSink *sw, char *data)
{   /* data is a string containing several new numbers to push into 
       the data.  See how many there are and incorporate them. */
    char *s = data;
    int i, n = 0, rc;
    double val;
    float *bptr;

    /* parse whole string once to count number of elements and 
       check syntax */
    while( (rc=findOneNumber(&s, &val)) == 1)  {
	++n;
    }
    if(rc < 0)  {
	Tcl_AppendResult(interp, "cannot parse as number :",
			 s, NULL);
	return TCL_ERROR;
    }
/*    fprintf(stderr, "parseData: %d data passed\n", n);   */
    /* OK, parse the string again, shift new data into our buffer */
    s = data;
    if(n > sw->bufLen)  {
	int discard = n - sw->bufLen;
#ifdef DEBUG
	fprintf(stderr, "parseData: discarding %d of %d data\n", discard, n);
#endif /* DEBUG */
	for(i=0; i<discard; ++i)
	    findOneNumber(&s, &val);	/* just parse over them */
	n -= discard;
    }
    /* How many of the existing buffer contents will be pushed out 
       by the new data?  As many as are required to respect bufLen */
    int toBeDropped  = max(0, n+sw->bufConts - sw->bufLen);
    /* so how many of the existing contents will we keep? */
    int goodOldConts = sw->bufConts - toBeDropped;
    /* what will the contents be after this filling? */
    int newBufConts = goodOldConts + n;
    /* The new items will go at the end of the buffer.  Copy back 
       the existing contents to make room, if necessary */
    if(toBeDropped > 0 && goodOldConts > 0) {
	DATAcopy(sw->buf+toBeDropped, sw->buf, goodOldConts);
	// VCopy(sw->buf+toBeDropped, 1, sw->buf, 1, goodOldConts);
    }
    bptr = sw->buf+goodOldConts;	/* base for new data */
    for(i=0; i<n; ++i)  {
	rc = findOneNumber(&s, &val);
	assert(rc==1);			/* since we already parsed it */
	*bptr++ = val;
    }
    sw->bufConts = newBufConts;
    return TCL_OK;
}

int wavesinkPushData(tkWaveSink *ws, DATA *d, int n)
{
    int totalDone = 0;

    if(n > ws->bufLen)  {
	int discard = n - ws->bufLen;
#ifdef DEBUG
	fprintf(stderr, "wsPushData: discarding %d of %d data\n", discard, n);
#endif /* DEBUG */
	d += discard;
	n -= discard;
    }
    /* How many of the existing buffer contents will be pushed out 
       by the new data?  As many as are required to respect bufLen */
    int toBeDropped  = max(0, n+ws->bufConts - ws->bufLen);
    /* so how many of the existing contents will we keep? */
    int goodOldConts = ws->bufConts - toBeDropped;
    /* what will the contents be after this filling? */
    int newBufConts = goodOldConts + n;
    /* The new items will go at the end of the buffer.  Copy back 
       the existing contents to make room, if necessary */
    if(toBeDropped > 0 && goodOldConts > 0) {
	DATAcopy(ws->buf+toBeDropped, ws->buf, goodOldConts);
//	VCopy(ws->buf+toBeDropped, 1, ws->buf, 1, goodOldConts);
    }
    float *bptr = ws->buf+goodOldConts;	/* base for new data */
    DATAcopy(d, bptr, (long)n);
//    VCopy(d, 1, bptr, 1, (long)n);
    ws->bufConts = newBufConts;
    return totalDone;
}

int WaveSink::Update(double /* time */ )
{   /* grab samples from the input */
    int totalDone = 0;
    if(ThereIsMore())  {
	assert(Input()->PopTail()>=0);
	int todo  = Input()->Avai();
	DATA *d;
	int n;
	while(todo)  {
	    n = min(todo, Input()->GetTailPtr(&d));
	    wavesinkPushData(ws, d, n);
	    Input()->UpdateTail(n);
	    todo -= n;
	    totalDone += n;
	}	
	if (!ws->updatePending) {
	    Tk_DoWhenIdle(wavesinkDisplay, (ClientData) ws);
	    ws->updatePending = 1;
	}
    }
    return totalDone;
}

static int wavesinkWidgetCmd(ClientData clientData, Tcl_Interp *interp, 
			    int argc, char **argv)
{   /* passed commands executed on the widget */
    tkWaveSink	*sw = (tkWaveSink *)clientData;
    int 	result = TCL_OK;
    int 	length;
    char	c;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " option ?arg arg ...?\"", NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData)sw);
    c = argv[1][0];
    length = strlen(argv[1]);
    if( (c=='c') && (strncmp(argv[1], "configure", length)==0) )  {
	if(argc==2) {
	    result = Tk_ConfigureInfo(interp, sw->tkwin, wavesinkConfigSpecs, 
				      (char *)sw, NULL, 0);
	} else if(argc==3) {
	    result = Tk_ConfigureInfo(interp, sw->tkwin, wavesinkConfigSpecs, 
				      (char *)sw, argv[2], 0);
	} else {
	    result = wavesinkConfigure(interp, sw, argc-2, argv+2, 
				      TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)) {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	result = Tk_ConfigureValue(interp, sw->tkwin, wavesinkConfigSpecs,
				   (char *) sw, argv[2], 0);
    } else if( (c == 'p') && (strncmp(argv[1], "push", length)==0) )  {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " push <datalist>\"", NULL);
	    return TCL_ERROR;
	}
        // argv3 is a bunch of numbers.  Maybe scan first
	result = wavesinkParseData(interp, sw, argv[2]);
    } else if( (c == 'd') && (strncmp(argv[1], "dump", length)==0) )  {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " dump\"", NULL);
	    return TCL_ERROR;
	}
	int i;
	char s[32];
	for(i=0; i<sw->bufConts; ++i)  {
	    sprintf(s, "%f ", sw->buf[i]);
	    Tcl_AppendResult(interp, s, (char *)NULL);
	}
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\":  must be configure, cget push or dump", (char *) NULL);
	result = TCL_ERROR;
    }
    if (result == TCL_OK && !sw->updatePending) {
	Tk_DoWhenIdle(wavesinkDisplay, (ClientData) sw);
	sw->updatePending = 1;
    }
    Tk_Release((ClientData)sw);
    return result;
}

static int wavesinkConfigure(Tcl_Interp *interp, tkWaveSink *ptr, 
			    int argc, char **argv, int flags)
{   /* handle configure commands */
    int rslt = TCL_OK;
    int origLen = ptr->bufLen;

    if(Tk_ConfigureWidget(interp, ptr->tkwin, wavesinkConfigSpecs, 
			  argc, argv, (char *)ptr, flags)!= TCL_OK)  {
	return TCL_ERROR;
    }
    // check for change of len
    if(ptr->bufLen != origLen)  {
	float *newbuf = (float *)ckalloc(ptr->bufLen*sizeof(float));
	if(ptr->buf)  {	/* i.e. there is some old data */
	    int tokeep;
	    tokeep = min(ptr->bufConts, ptr->bufLen);
	    // copy across data to keep
	    DATAcopy(newbuf, ptr->buf+ptr->bufConts-tokeep, tokeep);
	//    VCopy(newbuf, 1, ptr->buf+ptr->bufConts-tokeep, 1, tokeep);
	    ptr->bufConts = tokeep;
	    ckfree((char *)ptr->buf);
	}
	ptr->buf = newbuf;
    }
    // set the colors and GC for the redisplay
    Tk_SetWindowBackground(ptr->tkwin, Tk_3DBorderColor(ptr->bgBorder)->pixel);
    if (ptr->gc == None || ConfigSpecified("-linewidth", argc, argv)) {
	XGCValues gcValues;
	gcValues.function = GXcopy;
	gcValues.graphics_exposures = False;
	gcValues.line_width = ptr->lineWidth;
	// Tk maintains shareable, read-only graphics contexts.  So we 
	// can't change ours dynamically using X calls - someone else 
	// may be using it.  Rather, if we get a configuration request 
	// to change the line width, we free any existing GC and make 
	// a new one with the new GC.
	if(ptr->gc)	Tk_FreeGC(Tk_Display(ptr->tkwin), ptr->gc);
	ptr->gc = Tk_GetGC(ptr->tkwin,GCFunction|GCGraphicsExposures|GCLineWidth, &gcValues);
    }
    // register desired window geometry, arrange for redisplay
    Tk_GeometryRequest(ptr->tkwin, ptr->width, ptr->height);
    Tk_SetInternalBorder(ptr->tkwin, ptr->borderWidth);
    if (!ptr->updatePending) {
	Tk_DoWhenIdle(wavesinkDisplay, (ClientData)ptr);
	ptr->updatePending = 1;
    }
    // data sink part
    DataSource *in = 0;
    if(strcmp(ptr->input, NONESTR)!=0)  {
	// lookup the input in the hash table
	Tcl_HashEntry *entryPtr;
	entryPtr = Tcl_FindHashEntry(&datasrcTable, ptr->input);
	if(entryPtr == NULL)  {
	    Tcl_AppendResult(interp, "input \"", ptr->input, 
			     "\" does not match any data sources.", NULL);
	    return TCL_ERROR;
	} 
	in = (DataSource *)Tcl_GetHashValue(entryPtr);
    }
    if(ptr->obj == 0)  {
	if(in == 0 && !ConfigSpecified("-input", argc, argv))  {
	    in = globalDataSource;	// use the default - the last
	    ckfree(ptr->input);
	    ptr->input = strdup(globalDSnameStr);
	}
	ptr->obj = new WaveSink(ptr,in);
	// schedule it
	if(in != 0) {
	    if(globalScheduler==0)  {
		fprintf(stderr, "wavesinkConfigure:Error:no scheduler defined\n");
	    } else {
		globalScheduler->Enschedule(ptr->obj);
	    }
	}
    } else {
	DataSource *oldin = ptr->obj->Input();
	if(in != ptr->obj->Input())  {
	    ptr->obj->SetInput(in);
	    if(globalScheduler==0)  {
		fprintf(stderr, "wavesinkConfigure:Err:no scheduler def'd\n");
	    } else {
		if(in==0)  
		    globalScheduler->Deschedule(ptr->obj);
		else if(oldin==0)
		    globalScheduler->Enschedule(ptr->obj);
	    }
	}
    }	
    return rslt;
}

static void wavesinkEventProc(ClientData clientData, XEvent *eventPtr)
{				/* handle events for the WaveSink widget */
    tkWaveSink *sw = (tkWaveSink *)clientData;
    
    if(eventPtr->type == Expose)  {
	if(!sw->updatePending)  {
	    Tk_DoWhenIdle(wavesinkDisplay, (ClientData)sw);
	    sw->updatePending = 1;
	}
    } else if(eventPtr->type == ConfigureNotify)  {
	// adjust self for resize 
	if(!sw->updatePending)  {
	    Tk_DoWhenIdle(wavesinkDisplay, (ClientData)sw);
	    sw->updatePending = 1;
	}
    } else if(eventPtr->type == DestroyNotify)  {
	Tcl_DeleteCommand(sw->interp, Tk_PathName(sw->tkwin));
	sw->tkwin = NULL;
	if(sw->updatePending)  {
	    Tk_CancelIdleCall(wavesinkDisplay, (ClientData)sw);
	}
	Tcl_EventuallyFree((ClientData)sw, (Tcl_FreeProc *)wavesinkDestroy);
    }
}

static void wavesinkDrawWvfm(tkWaveSink *sw, Drawable d=None)
{   /* actually draw WaveSink data from our buf */
    Tk_Window 		tkwin = sw->tkwin;
    GC 			gc    = sw->gc;
    Display 		*disp = Tk_Display(tkwin);
    XColor 		*fgColor = sw->fgColor;
    double 		scale   = sw->scale;
    int x, y, w, h, y0, yscale, xd=0, yd=0, xo, yo;
    int i;
    int nxs, xssize = 100;
    XSegment *xsgs, *thisSeg;
    float	*bptr = sw->buf;
    int		conts= sw->bufConts;
    int		len  = sw->bufLen;
    int 	offs = len-conts;
    
    if(d==None)
	d = Tk_WindowId(tkwin);
    
    wavesinkGetWorkArea(sw, &x, &y, &w, &h);

    yscale = -h/2;
    y0 = y+h/2 + (int)(sw->offset*yscale);
    
    XSetForeground(disp, gc, fgColor->pixel);
    
    xsgs = (XSegment *)ckalloc(xssize * sizeof(XSegment));
    thisSeg = xsgs;
    nxs = 0;
    for(i = 0; i < conts; ++i)  {
	yo = yd;  xo = xd;
	yd = int(y0+yscale*scale**bptr++);
	if(yd >= y+h)	yd = y+h-1;
	if(yd < y)	yd = y;
	xd = x+(((offs+i)*w)/(len-1));	// len-1 because that's the # of segs
	if(i>0)  {
	    thisSeg->x1 = xo;	thisSeg->y1 = yo;
	    thisSeg->x2 = xd;	thisSeg++->y2 = yd;
	    if(++nxs == xssize)  {
		XDrawSegments(disp, d, gc, xsgs, nxs);
		thisSeg = xsgs;
		nxs = 0;
	    }
	}
    }
    if(nxs > 0)  {
	XDrawSegments(disp, d, gc, xsgs, nxs);
    }
    ckfree((char *)xsgs);
}

static void wavesinkDisplay(ClientData clientData)
{				/* actually draw the contents of the widget */
    tkWaveSink *sw = (tkWaveSink *)clientData;
    Tk_Window tkwin = sw->tkwin;
    Pixmap pm = None;
    Drawable d;
    Display *disp = sw->display;
    int 	tw, th, inset;
    
    sw->updatePending = 0;
    if(!Tk_IsMapped(tkwin)) {
	return;
    }
    
    tw = Tk_Width(tkwin);
    th = Tk_Height(tkwin);
    
    // Create pixmap for double-buffering if necessary
    if(sw->doubleBuffer)  {
	pm = XCreatePixmap(disp, Tk_WindowId(tkwin), tw, th, 
			   DefaultDepthOfScreen(Tk_Screen(tkwin)));
	d = pm;
    } else {		// draw directly into window
	d = Tk_WindowId(tkwin);
    }
    // redraw background
    Tk_Fill3DRectangle(tkwin, d, sw->bgBorder, 0, 0, tw, th, sw->borderWidth, 
		       sw->relief);
    // Actually do the drawing
    inset = 2;
    
    //    drawX(tkwin, sw->gc, d, sw->borderWidth + inset);
    wavesinkDrawWvfm(sw, d);
    
    // copy to screen if double-buffering
    if(sw->doubleBuffer)  {
	XCopyArea(disp, pm, Tk_WindowId(tkwin), sw->gc, 0, 0, 
		  tw, th, 0, 0);
	XFreePixmap(disp, pm);
    }
}

static void wavesinkDestroy(char* clientData)
{				/* destroy one of our objects */
    tkWaveSink *sw = (tkWaveSink *)clientData;
    
    Tk_FreeOptions(wavesinkConfigSpecs, (char *)sw, sw->display, 0);
    if(sw->gc != None)  {
	Tk_FreeGC(sw->display, sw->gc);
    }
    if(sw->obj != 0)
	delete sw->obj;		// the WaveSink object
    if(sw->buf != 0)
	ckfree((char *)sw->buf);
    ckfree((char *)sw);
}


