/*
 * tkImgFmtFarray.C --   (after tk4.0/tkImgFmtPPM.c)
 *
 *	A photo image handler for floatarray tcl objects.
 * 	floatArrays are ObjectTcl versions of a C++ template class
 *	in Array.H
 *
 * dpwe@media.mit.edu 1995jul07
 */

// Only do something if we were configured with floatArrays

#ifdef HAVE_FARRAY

#include <iostream.h>

extern "C" {
#include <tkInt.h>
#include <tkPort.h>
}

/* 
 * Local includes 
 */
#include <floatArray/floatArray.H>	/* definitions of array of floats */
//#include "nSurface.H" /* instead of floatArray.H */
//#include <Otcl.H>	/* definitions of ObjectTcl interface */
#include <floatArray/floatArrayTcl.H>	// declare FAfromName()

static int		StringMatchFA(char *string,
			    char *formatString, int *widthPtr,
			    int *heightPtr);
static int		StringReadFA(Tcl_Interp *interp,
			    char *string, char *formatString,
			    Tk_PhotoHandle imageHandle, int destX, int destY,
			    int width, int height, int srcX, int srcY);
static int		StringWriteFA(Tcl_Interp *interp,
			    Tcl_DString *string, char *formatString,
			    Tk_PhotoImageBlock *blockPtr);

/*
 * The format record for the floatArray format:
 */

Tk_PhotoImageFormat tkImgFmtFA = {
    "floatArray",		/* name */
    NULL,			/* fileMatchProc */
    StringMatchFA,		/* stringMatchProc */
    NULL,			/* fileReadProc */
    StringReadFA,		/* stringReadProc */
    NULL,			/* fileWriteProc */
    StringWriteFA,		/* stringWriteProc */
    NULL,		
};

#define MAX_MEMORY 100000

/*
 * Prototypes for local procedures defined in this file:
 */

/* colormap handling */

static int FAimg_ncolors = 0;
static int *FAimg_cmap = NULL;

/* built-in colormaps */

typedef struct _cmap {
    struct _cmap *next;
    char *name;
    int  ncolors;
    int  *rgbs;
} CMAP;

int tidata[] = 
{0, 175, 175, 0, 169, 175, 0, 163, 175, 0, 157, 175, 0, 151, 175, 
0, 145, 175, 0, 139, 175, 0, 133, 175, 0, 121, 175, 0, 115, 175, 
0, 109, 175, 0, 103, 175, 0, 97, 175, 0, 91, 175, 0, 85, 175, 
0, 79, 175, 0, 67, 175, 0, 61, 175, 0, 55, 175, 0, 49, 175, 
0, 43, 175, 0, 37, 175, 0, 31, 175, 0, 0, 128, 4, 0, 124, 
6, 0, 122, 8, 0, 120, 10, 0, 118, 12, 0, 116, 14, 0, 114, 
16, 0, 112, 18, 0, 110, 22, 0, 106, 24, 0, 104, 26, 0, 102, 
28, 0, 100, 30, 0, 98, 32, 0, 96, 34, 0, 94, 36, 0, 92, 
40, 0, 88, 42, 0, 86, 44, 0, 84, 46, 0, 82, 48, 0, 80, 
64, 0, 78, 70, 0, 76, 76, 0, 74, 88, 0, 70, 94, 0, 68, 
100, 0, 66, 106, 0, 64, 112, 0, 62, 118, 0, 60, 124, 0, 58, 
130, 0, 56, 142, 0, 52, 148, 0, 50, 154, 0, 48, 160, 0, 46, 
166, 0, 44, 172, 0, 42, 178, 0, 40, 184, 0, 38, 196, 0, 34, 
202, 0, 32, 208, 0, 30, 255, 0, 0, 255, 10, 0, 255, 20, 0, 
255, 30, 0, 255, 40, 0, 255, 60, 0, 255, 70, 0, 255, 80, 0, 
255, 90, 0, 255, 100, 0, 255, 110, 0, 255, 120, 0, 255, 130, 0, 
255, 150, 0, 255, 160, 0, 255, 170, 0, 255, 180, 0, 255, 190, 0, 
255, 200, 0, 255, 210, 0, 255, 220, 0, 255, 240, 0, 255, 255, 0, 
255, 255, 8, 255, 255, 16, 255, 255, 24, 255, 255, 32, 255, 255, 40, 
255, 255, 48, 255, 255, 64, 255, 255, 72, 255, 255, 80, 255, 255, 88, 
255, 255, 96, 255, 255, 104, 255, 255, 112, 255, 255, 120, 255, 255, 136, 
255, 255, 144, 255, 255, 152, 255, 255, 160, 255, 255, 168, 255, 255, 176, 
255, 255, 184, 255, 255, 192, 255, 255, 208, 255, 255, 216, 255, 255, 224};

CMAP TImap = {NULL, "esps", 115, tidata};
CMAP nonemap = {&TImap, "none", 0, NULL};

CMAP *cmap_root = &nonemap;

int findNamedCmap(char *name) {
    /* search through the linked list of known cmaps, maybe select one */
    CMAP *cmap_ptr = cmap_root;
    int i, found = 0;

    while(!found && cmap_ptr != NULL) {
	// fprintf(stderr, "findNamedCmap: checking '%s' against '%s'...\n", name, cmap_ptr->name);
	if (strcmp(name, cmap_ptr->name)==0) {
	    /* we found a match */
	    found = 1;
	    /* copy it to our current cmap structure */
	    FAimg_ncolors = cmap_ptr->ncolors;
	    if (FAimg_cmap != NULL) {
		ckfree((char *)FAimg_cmap);
		FAimg_cmap = NULL;
	    }
	    if (FAimg_ncolors > 0) {
		FAimg_cmap = (int *)ckalloc(3*FAimg_ncolors*sizeof(int));
		for (i = 0; i < 3*FAimg_ncolors; ++i) {
		    FAimg_cmap[i] = cmap_ptr->rgbs[i];
		}
	    }
	}
	cmap_ptr = cmap_ptr->next;
    }
    return found;
}

extern "C" int
tclFAimgSetCmapCmd(ClientData cdata, Tcl_Interp *interp, 
		   int argc, char *argv[])
{   /* command to set the colormap used in displaying tclFAimgs */
    /* farraycmap [list] , null list to revert to b/w map */
    int myargc, myargc2;
    char **myargv, **myargv2;
    int *newcmap = NULL;
    int err = 0;
    int i, j, clr;

    if (argc > 2) {
	Tcl_AppendResult(interp, "wrong # args: \"", argv[0], " ?cmaplist?\"", 
			 NULL);
	return TCL_ERROR;
    }
    /* the list should be a list of three-element int lists */

    if (argc < 2) {
	/* no argument -> de-install colormap */
	myargc = 0; 
    } else {
	if ( Tcl_SplitList(interp, argv[1], &myargc, &myargv) != TCL_OK ) {
	    err = 1;
	}
    }
    if (err == 0) {
	if (myargc == 1) {
	    /* single argument means colormap name - search list */
	    if (!findNamedCmap(myargv[0])) {
		Tcl_AppendResult(interp, "unknown cmap name \"", 
				 myargv[0], "\"", NULL);
		/* bail now */
		Tcl_Free((char *)myargv);
		return TCL_ERROR;
	    }
	} else {
	    if (myargc > 0) {
		/* argument was a list - use it to allocate the new cmap */
		newcmap = (int *)ckalloc(3*myargc*sizeof(int));
		/* decode the colormap list */
		for(i=0; i<myargc; ++i) {
		    if ( Tcl_SplitList(interp, myargv[i], &myargc2, &myargv2) \
			 != TCL_OK ) {
			err = 1;
			break;
		    }
		    if (myargc2 != 3) {
			err = 1;
			break;
		    }
		    for (j=0; j<3; ++j) {
			if(Tcl_GetInt(interp, myargv2[j], &clr) != TCL_OK) {
			    err = 1;
			    break;
			}
			newcmap[3*i+j] = clr;
		    }
		    if(myargc2) Tcl_Free((char*)myargv2);
		}
	    }
	    if (err == 0) {
		FAimg_ncolors = myargc;
		if (FAimg_cmap != NULL) {
		    ckfree((char *)FAimg_cmap);
		    FAimg_cmap = NULL;
		}
		if (FAimg_ncolors > 0) {
		    FAimg_cmap = newcmap;
		}
	    }
	}
	if(myargc)  Tcl_Free((char*)myargv);
    }
    if (err) {
	if (newcmap)	free(newcmap);
	Tcl_AppendResult(interp, argv[0], ": cmaplist should be a known name or list of 3-integer lists", NULL);
	return TCL_ERROR;
    }
    sprintf(interp->result, "%d", FAimg_ncolors);
    return TCL_OK;
}


extern "C" int ImgFmtFA_Init(Tcl_Interp *interp) {
    // Call the fn to get this photo type registered
    Tk_CreatePhotoImageFormat(&tkImgFmtFA);
    // Create the farraycmap command
    Tcl_CreateCommand( interp, "floatArrayCmap", tclFAimgSetCmapCmd, 
                       (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL );
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GetFarrayObj
 *
 * 	Translate a string into a pointer to a floatArray, using the 
 *	hash table defined by the ObjectTcl definition of floatArray objs.
 *
 * Results:
 *	a pointer to a floatArray object, or NULL if the name is invalid.
 *
 * Side effects:
 *	none
 *
 *----------------------------------------------------------------------
 */

static floatArray*
GetFarrayObj(
    char *string)		/* The string to match. */
{
    floatArray *faPtr;

    faPtr = FAfromName(string);
    if(faPtr == NULL) {  // no floatArray registered to that name
	return NULL;
    }
    return faPtr;
}


/*
 *----------------------------------------------------------------------
 *
 * StringMatchFA --
 *
 *	This procedure is invoked by the photo image type to see if
 *	a given string can be mapped to a floatArray image i.e. if it 
 *	is the name of a floatArray
 *
 * Results:
 *	The return value is >0 if the <string> exists in the floatArray
 *	hash table, and 0 otherwise.
 *
 * Side effects:
 *	If the floatArray is found, its size is entered into 
 *	<*widthPtr> and <*heightPtr>
 *
 *----------------------------------------------------------------------
 */

static int
StringMatchFA(
    char *string,		/* The string to match. */
    char *formatString,		/* User-specified format string, or NULL. */
    int *widthPtr,
    int *heightPtr)		/* The dimensions of the image are
				 * returned here if the file is a valid
				 * raw PPM file. */
{
    floatArray *faPtr;

    if((faPtr=GetFarrayObj(string))!=NULL)  {
	*widthPtr = faPtr->Cols();
	*heightPtr = faPtr->Rows();
	return 1;
    } else {
	return 0;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * StringReadFA --
 *
 *	This procedure is called by the photo image type to read
 *	the data from a floatArray object and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	New data is added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

static void
PhotoFromFA_nocmap(
    floatArray *faPtr,
    Tk_PhotoHandle imageHandle,	/* The photo image to write into. */
    int srcX, 
    int srcY,			/* Coordinates of top-left pixel in
				 * array to read from. */
    int destX, 
    int destY,			/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, 
    int height)			/* Dimensions of block of photo image to
				 * be written to. */
{
    int arWidth, arHeight;
    int nLines, nBytes, h;
    unsigned char *pixelPtr;
    Tk_PhotoImageBlock block;

    block.pixelSize = 1;
    block.offset[0] = 0;
    block.offset[1] = 0;
    block.offset[2] = 0;
    block.width = width;
    block.pitch = block.pixelSize * width;

    nLines = (MAX_MEMORY + block.pitch - 1) / block.pitch;
    if (nLines > height) {
	nLines = height;
    }
    if (nLines <= 0) {
	nLines = 1;
    }
    nBytes = nLines * block.pitch;
    pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
    block.pixelPtr = pixelPtr;

    int x, y, d;
    unsigned char *ppTmp;

    for (h = height; h > 0; h -= nLines) {
	if (nLines > h) {
	    nLines = h;
	}
	ppTmp = pixelPtr;
	for(y = 0; y<nLines; ++y)  {
	    for(x = 0; x<width; ++x)  {
                d = (int)(0.5 + 255.0 * faPtr->El(srcY+h-y-1, srcX+x));
		if(d < 0) 	d = 0;
		if(d > 255)	d = 255;
		/* default scaling is 0->white, FSD->black ... */
		*ppTmp++ = (255-d);
	    }
	}
	block.height = nLines;
	Tk_PhotoPutBlock(imageHandle, &block, destX, destY, width, nLines);
	destY += nLines;
    }

    ckfree((char *) pixelPtr);
}

static void
PhotoFromFA_cmap(
    floatArray *faPtr,
    Tk_PhotoHandle imageHandle,	/* The photo image to write into. */
    int srcX, 
    int srcY,			/* Coordinates of top-left pixel in
				 * array to read from. */
    int destX, 
    int destY,			/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, 
    int height,			/* Dimensions of block of photo image to
				 * be written to. */
    int rgbcolormap[],		/* colormap to set pseudocolors */
    int ncolors)
{
    int arWidth, arHeight;
    int nLines, nBytes, h;
    unsigned char *pixelPtr;
    Tk_PhotoImageBlock block;

    block.pixelSize = 3;
    block.offset[0] = 0;
    block.offset[1] = 1;
    block.offset[2] = 2;
    block.width = width;
    block.pitch = block.pixelSize * width;

    nLines = (MAX_MEMORY + block.pitch - 1) / block.pitch;
    if (nLines > height) {
	nLines = height;
    }
    if (nLines <= 0) {
	nLines = 1;
    }
    nBytes = nLines * block.pitch;
    pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
    block.pixelPtr = pixelPtr;

    int x, y, d;
    unsigned char *ppTmp;

    for (h = height; h > 0; h -= nLines) {
	if (nLines > h) {
	    nLines = h;
	}
	ppTmp = pixelPtr;
	for(y = 0; y<nLines; ++y)  {
	    for(x = 0; x<width; ++x)  {
		/* calculate index into colormap */
                d = (int)((ncolors+1.0) * faPtr->El(srcY+h-y-1, srcX+x));
		if(d < 0) 	d = 0;
		if(d >= ncolors)	d = ncolors-1;
		/* copy across r,g,b values */
		*ppTmp++ = rgbcolormap[3*d];
		*ppTmp++ = rgbcolormap[3*d+1];
		*ppTmp++ = rgbcolormap[3*d+2];
	    }
	}
	block.height = nLines;
	Tk_PhotoPutBlock(imageHandle, &block, destX, destY, width, nLines);
	destY += nLines;
    }

    ckfree((char *) pixelPtr);
}

static int
StringReadFA(
    Tcl_Interp *interp,		/* Interpreter to use for reporting errors. */
    char *string,		/* The name of the floatArray object. */
    char *formatString,		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle,	/* The photo image to write into. */
    int destX, 
    int destY,			/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, 
    int height,			/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, 
    int srcY)			/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
    int arWidth, arHeight;
    int nLines, nBytes, h;
    unsigned char *pixelPtr;
    Tk_PhotoImageBlock block;

    floatArray *faPtr;

    faPtr = GetFarrayObj(string);
    if (faPtr == NULL) {
	Tcl_AppendResult(interp, "couldn't find floatArray \"",
		string, "\"", NULL);
	return TCL_ERROR;
    }
    arWidth = faPtr->Cols();
    arHeight = faPtr->Rows();
    if ((arWidth <= 0) || (arHeight <= 0)) {
	Tcl_AppendResult(interp, "floatArray \"", string,
		"\" has dimension(s) <= 0", (char *) NULL);
	return TCL_ERROR;
    }

    if ((srcX + width) > arWidth) {
	width = arWidth - srcX;
    }
    if ((srcY + height) > arHeight) {
	height = arHeight - srcY;
    }
    if ((width <= 0) || (height <= 0)
	|| (srcX >= arWidth) || (srcY >= arHeight)) {
	return TCL_OK;
    }
    Tk_PhotoExpand(imageHandle, destX + width, destY + height);

    if (FAimg_ncolors == 0) {
	PhotoFromFA_nocmap(faPtr, imageHandle, srcX, srcY, destX, destY, width, height);
    } else {
	PhotoFromFA_cmap(faPtr, imageHandle, srcX, srcY, destX, destY, width, height, 
			 FAimg_cmap, FAimg_ncolors);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * StringWriteFA --
 *
 *	This procedure is invoked to translate image data into a 
 *	new floatArray object.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	A new floatArray object is created, filled with the data in the 
 *	photo object, and its Tcl name is returned in <dataPtr>.
 *
 *----------------------------------------------------------------------
 */

static int
StringWriteFA(
    Tcl_Interp *interp, 
    Tcl_DString *string, 
    char *formatString, 
    Tk_PhotoImageBlock *blockPtr)
{
    Tcl_AppendResult(interp, "no 'put' supported for floatArray", NULL);
    return TCL_ERROR;
}

#endif /* HAVE_FARRAY */
