#
# thislnlp.tcl
#
# Interface the thislgui demo to the Thomson natural language parsing (NLP)
# stuff.
#
# 1999may04 dpwe@icsi.berkeley.edu 
# $Header: /u/drspeech/src/thislgui/RCS/thislnlp.tcl,v 1.3 2000/03/22 02:54:30 dpwe Exp $

# Where the runparser binary is
#set ParserDir "/u/dpwe/tmp/thisl/thomson/nathalie/bin"
# Where the parse file appears
#set NLPRSLTPATH "/u/dpwe/tmp/thisl/thomson/nathalie/src"

# Static counts utterances, to find right lattice file
set nlp_uttcount 0

proc getTerminals {tree {typepat ".*"}} {
    # Take a {TYPE child1 child2}-style parse tree and return 
    # just the terminals in a flat string
    # Only return terminals whose type matches the regexp $typepat
    # - by default, all of them
    set type [lindex $tree 0]
    set childlist [lrange $tree 1 e]
    if {$childlist == ""} {
	# Singleton - fake a pair
	set childlist $type
	set type FAKE
    }
    if {[llength [lindex $childlist 0]] == 1} {
	# This is a terminal - return it, if the type is good
	if {[regexp $typepat $type]} {
	    set rslt $childlist
	} else {
	    set rslt ""
	}
    } else {
	# Concatenate the terminals of all the children
	set rslt ""
	foreach child $childlist {
	    set rslt [concat $rslt [getTerminals $child $typepat]]
	}
    }
    return $rslt
}

proc NLPparse {line} {
    # Read the next lattice and return best word result
    global latticebase nlp_uttcount ParserDir
# NLPRSLTPATH
    global treecanv decodetext parsedtext keywdtext

    Warn "NLPparse $line"
    # Clear out previous displays
    NLPclear
    # Write up decoder output
    $decodetext config -text [string tolower $line]

    incr nlp_uttcount
    set latfile "${latticebase}.$nlp_uttcount"
    if {![file readable $nlp_uttcount]} {
	Warn "pausing for '$latfile' to appear..."
	exec sleep 1
    }
    # Assume it's there now
#    afe_setstatus "parsing..."; update
    STATUSset "parsing..."; update
    set rslt [string trim [exec $ParserDir/runparser $latfile 2>@stderr]]
    # output value is actually the best hyp
    # no, output value is now the parse tree
    Warn "NLPparse gave '$rslt'"

    # Read parse tree file
#    set parsefile "$NLPRSLTPATH/first_gram_analysis.pl"
#    set f [open $parsefile  "r"]
#    set parse [read $f]
#    close $f
    # No, parse tree text is now return from runparser
    set parse $rslt

    # Map to a tcl list
    # .. remove trash at start
    regsub {parsed\(} $parse "" parse
    # .. maybe a trailing paren if one ever shows up
    regsub {\)$} $parse "" parse
    # .. map sq brackets to Tcl braces
    regsub -all {\[} $parse \{ parse
    regsub -all {\]} $parse \} parse
    # .. map commas to space delimiters
    regsub -all {,} $parse " " parse
    # .. remove any args inside parens (to avoid breaks)
    regsub -all {\([^)]*\)} $parse "" parse
    # Strip a level of braces (or abandon if bad return)
    if {[catch {set parse [lindex $parse 0]}]} {
	# $parse was not a valid list
	Warn "Garbled return from runparser, skipping"
	set rslt $line
    } else {
	# Add a root node
	set parse [concat "S" $parse]
	# Report result to console
	Warn "parse= $parse"

	# Convert to the single best hyp & display
	set kpat {^(k|unk)$}
	set rslt [getTerminals $parse]
	set kwds [getTerminals $parse $kpat]
	$parsedtext config -text [string tolower $rslt]
	$keywdtext config -text [string tolower $kwds]

	# If keyword hits, use instead of whole thing
	if {$kwds != ""} {
	    set rslt $kwds
	}

	# Draw the parse tree
    #    DrawTreeVert $treecanv $parse 10 10
	DrawTreeHoriz $treecanv $parse 10 10 black $kpat
    }

    # Clean up - remove the lat file
    file delete $latfile

    return $rslt
}

#
# parsetree.tcl
# 
# Draw a parse tree
# 
# 1999may05 dpwe@icsi.berkeley.edu
# $Header: /u/drspeech/src/thislgui/RCS/thislnlp.tcl,v 1.3 2000/03/22 02:54:30 dpwe Exp $

set tree {S {NP {A the} {N man}} {VP {V looked} {PP {P at} {NP {A the} {N cat}}}}}

proc DrawTreeVert {canv tree x y} {
    # Recusive tree-drawing.  x, y are top-left of allocated rectangle. 
    # Return the x of the right edge of tree below
    # Iterate over children drawing them in
    global fontArray
    # Row height (offset to next row)
    set height 40
    # Spacing between adjacent children
    set xgap 10
    # Gap above text
    set ygap 3
    
    set ychild [expr $y + $height]
    set xleft [expr $x + $xgap/2]
    set type [lindex $tree 0]
    set children [lrange $tree 1 e]
    # Choose font based on being a terminal or not
    if {[llength $children] == 0} {
	set font $fontArray(helv12b)
    } else {
	set font $fontArray(helv12)
    }
    # List of points to connect edges to
    set childtops {}
    set xchild $x
    foreach child $children {
	# Each child abuts predecessor
	set lastxc $xchild
	# Skip null children
	if {$child != ""} {
	    set xchild [DrawTreeVert $canv $child $xchild $ychild]
	    lappend childtops [expr ($xchild+$lastxc)/2]
	}
    }
    # Now figure center for drawing self
    set myx [expr ($x + $xchild)/2]
    # Draw self
    set self [$canv create text $myx [expr $y+$ygap] -anchor n -text $type -font $font]
    set bbox [$canv bbox $self]
    # If our left edge is too far left, scoot right
    set ledge [lindex $bbox 0]
    set redge [lindex $bbox 2]
    if {$ledge < $xleft} {
	set xshift [expr $xleft - $ledge]
	$canv move $self $xshift 0
	incr ledge $xshift
	incr redge $xshift
    }
    # Figure center of bottom for edges
    set ybase [lindex $bbox 3]
    set xmid [expr ($ledge+$redge)/2]
    foreach top $childtops {
	$canv create line $xmid $ybase $top $ychild
    }
    # Return val is right edge (or extent of label)
    set retval [expr $redge + $xgap/2]
    if {$xchild > $retval} {
	set retval $xchild
    }
    return $retval
}

proc DrawTreeHoriz {canv tree x y {color "black"} {typepat "NONE"}} {
    # Recusive tree-drawing.  x, y are top-left of allocated rectangle. 
    # Return the x of the right edge of tree below
    # Iterate over children drawing them in
    # ** LEFT_TO_RIGHT version
    global fontArray

    # Gap between lines
    set ygap 6

    # Gap between columns
    set hgap 20

    # Head of list is this node's type, remainder is children
    set type [lindex $tree 0]
    set children [lrange $tree 1 e]

    # Choose font based on being a terminal or not
    if {[llength $children] == 0} {
	set font $fontArray(helv12b)
    } else {
	set font $fontArray(helv12)
    }

    # Choose color based on typematch
    if {[regexp $typepat $type]} {
	set color "red"
    }

    # Draw self first to get width
    set self [$canv create text $x [expr $y + $ygap/2] \
	    -anchor nw -text $type -font $font -fill $color]
    lassign [$canv bbox $self] bxl byt bxr byb

    # Children beyond the end
    set xchild [expr $bxr + $hgap]
    set ychild $y
    
    # List of points to connect edges to
    set childtops {}
    foreach child $children {
	# Each child abuts predecessor
	set lastyc $ychild
	# Skip null children
	if {$child != ""} {
	    set ychild [DrawTreeHoriz $canv $child $xchild $ychild \
		    $color $typepat]
	    lappend childtops [expr ($ychild+$lastyc)/2]
	}
    }

    # Now figure center for drawing self
    if {$ychild > $y} {
	set desy [expr ($y + $ychild)/2]
	# Adjust self to center rel. to children
	set yshift [expr $desy - ($byt+$byb)/2]
	$canv move $self 0 $yshift
	incr byt $yshift
	incr byb $yshift
    }

    # Draw connecting edges
    set ymid [expr ($byt+$byb)/2]
    foreach top $childtops {
	$canv create line $bxr $ymid $xchild $top -fill $color
    }

    # Return val is bottom of our label
    set retval [expr $byb + $ygap/2]
    # .. or lowest point of children
    if {$ychild > $retval} {
	set retval $ychild
    }

    return $retval
}

set treecanv ""

proc NLPmake {p} {
    # Create the canvas for later tree display inside $path
    global treecanv decodetext parsedtext keywdtext
    set pad 5
    frame $p -relief groove -borderwidth 2

    # Lines at the top for input and output hyps
    set t1 [frame $p.top1]
    label $t1.label -text "Recog:" -width 8
    set decodetext [label $t1.val -relief sunken -borderwidth 1 -text "" -anchor w]
    pack $t1.label -side left -expand 0 -fill y
    pack $t1.val -side left -expand 1 -fill both

    # .. and for parse output
    set t2 [frame $p.top2]
    label $t2.label -text "Parsed:" -width 8
    set parsedtext [label $t2.val -relief sunken -borderwidth 1 -text "" -anchor w]
    pack $t2.label -side left -expand 0 -fill y
    pack $t2.val -side left -expand 1 -fill both

    # .. and for keywords
    set t3 [frame $p.top3]
    label $t3.label -text "Keywds:" -width 8
    set keywdtext [label $t3.val -relief sunken -borderwidth 1 -text "" -anchor w]
    pack $t3.label -side left -expand 0 -fill y
    pack $t3.val -side left -expand 1 -fill both

    # pack the containers
    pack $t1 $t2 $t3 -side top -fill x -expand 0 -padx $pad -pady $pad

    set treecanv $p.canv
    #canvas $treecanv
    #pack $treecanv -fill both -expand 1
    
    # Scrollable canvas below
    set w $p
    set c $treecanv
    frame $w.grid
    scrollbar $w.hscroll -orient horiz -command "$c xview"
    scrollbar $w.vscroll -command "$c yview"
    canvas $c -relief sunken -borderwidth 2 -scrollregion {0c 0c 20c 20c} \
	-xscrollcommand "$w.hscroll set" \
	-yscrollcommand "$w.vscroll set"
    pack $w.grid -expand yes -fill both -padx $pad -pady $pad
    grid rowconfig    $w.grid 0 -weight 1 -minsize 0
    grid columnconfig $w.grid 0 -weight 1 -minsize 0

    grid $c -padx 1 -in $w.grid -pady 1 \
	    -row 0 -column 0 -rowspan 1 -columnspan 1 -sticky news
    grid $w.vscroll -in $w.grid -padx 1 -pady 1 \
	    -row 0 -column 1 -rowspan 1 -columnspan 1 -sticky news
    grid $w.hscroll -in $w.grid -padx 1 -pady 1 \
	    -row 1 -column 0 -rowspan 1 -columnspan 1 -sticky news

}
    
proc NLPclear {} {
    # Clear all the displays in the NLP panel
    global treecanv decodetext parsedtext keywdtext
    $treecanv delete all
    $decodetext config -text ""
    $parsedtext config -text ""
    $keywdtext config -text ""
}
